from loguru import logger
import json
from converters import CONVERTERS
from dataclasses import dataclass, field, asdict

@dataclass(init=True, kw_only=True)
class Context:
    high_temperature: str
    general_off: bool = field(default=False)
    maintenance_mode: bool = field(default=False)
    overwrite_window: bool = field(default=False)
    window_state: dict = field(default_factory=dict)
    feedback: dict = field(default_factory=dict)
    mode: str = field(default='off')
    output_temperature: str = field(default='0')

    def __str__(self):
        return json.dumps(asdict(self))

class Box:
    def __init__(self, box_id, box_config, config):
        logger.info(f"[Box {box_id}] Instantiating")

        self.id = box_id
        self.windows = box_config['windows']
        self.output_converter = box_config['output_converter']
        self.output_topic = box_config['output_topic']
        # we use get here since this key is optional
        self.feedback_topic = box_config.get('feedback_topic')
        self.config = config

        self.context = Context(high_temperature=config.DEFAULT_HIGH_TEMPERATURE,
                               output_temperature=config.DEFAULT_HIGH_TEMPERATURE,
                               mode='high',
                               window_state={ k: 'closed' for k in self.windows.keys() })

        self.mqtt_client = None

        logger.info(f"[Box {box_id}] Instantiated, context is {self.context}")

    def _calculate_output_temperature(self):
        # maintenance_mode has the highest priority, even higher than general_off
        if self.context.maintenance_mode:
            self.context.output_temperature = self.config.MAINTENANCE_TEMPERATURE
            return
        # general_off has the next highest priority
        if self.context.general_off:
            self.context.output_temperature = self.config.OFF_TEMPERATURE
            return
        # an open window shuts off the heating
        if not self.context.overwrite_window:
            for v in self.context.window_state.values():
                if v == 'open':
                    self.context.output_temperature = self.config.OFF_TEMPERATURE
                    return
        # finally evaluate the mode
        if self.context.mode == 'off':
            self.context.output_temperature = self.config.OFF_TEMPERATURE
            return
        if self.context.mode == 'low':
            self.context.output_temperature = self.config.LOW_TEMPERATURE
            return
        if self.context.mode == 'high':
            self.context.output_temperature = self.context.high_temperature
            return
        # if we come here, something serious happened
        logger.error(f"Error in calculation of output_temperature: {self.context=}")
        return


    def handle_message(self, topic_key, payload):
        logger.info(f"[Box {self.id}] Handle message for '{topic_key}': {payload}")

        try:
            # match topic to find operation to be executed
            match topic_key.split('/'):
                case [ primary_key, sub_key ] if primary_key == 'window':
                    self.context.window_state[sub_key] = CONVERTERS["window_contact_input"][self.windows[sub_key]["converter"]](payload)
                case [ primary_key ] if primary_key == 'high_temp':
                    self.context.high_temperature = payload
                case [ primary_key ] if primary_key == 'cmd':
                    p = payload.lower()
                    if p in ('high', 'low', 'off'):
                        self.context.mode = p
                    else:
                        raise Exception(f"Invalid cmd '{payload}'")
                case [ primary_key ] if primary_key == 'overwrite_window':
                    self.context.overwrite_window = payload.lower() == 'true'
                case [ primary_key ] if primary_key == 'general_off':
                    self.context.general_off = payload.lower() == 'true'
                case [ primary_key ] if primary_key == 'maintenance_mode':
                    self.context.maintenance_mode = payload.lower() == 'true'
                case [ primary_key ] if primary_key == 'status':
                    pass
                case [ primary_key ] if primary_key == 'feedback':
                    self.context.feedback = json.loads(payload)
                case _:
                    raise Error(f"Unexcepted topic_key: {topic_key}, {payload}")

            # calculate output temperature from context
            self._calculate_output_temperature()


            # publish output temperature
            result_message = CONVERTERS["target_temperature_output"][self.output_converter](self.context.output_temperature)
            publish_topic = self.output_topic
            self.mqtt_client.publish(publish_topic, result_message)
            logger.info(f"[Box {self.id}] Result published on '{publish_topic}': {result_message}")

            context_topic = f"{self.config.CONTEXT_TOPIC_PREFIX}{self.id}"
            self.mqtt_client.publish(context_topic, str(self.context))
        except Exception as e:
            logger.error(f"[Box {self.id}] Error processing '{topic_key}': {e}")
